﻿using System;
using System.Diagnostics;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  /// <summary>
  /// An algebraic type which has a value as well as the partial derivatives of the value over multiple variables.
  /// </summary>
  /// <typeparam name="V"></typeparam>
  [DebuggerDisplay("v={Value}; dv={dv}")]
  public class MultivariateDual<V> : IAlgebraicType<MultivariateDual<V>> where V : IAlgebraicType<V>, new() {
    [DebuggerBrowsable(DebuggerBrowsableState.Never)]
    private V v;
    public V Value => v;

    [DebuggerBrowsable(DebuggerBrowsableState.Never)]
    private AlgebraicSparseVector<object, V> dv;
    public AlgebraicSparseVector<object, V> Gradient => dv; // <key,value> partial derivative identified via the key

    private MultivariateDual(MultivariateDual<V> orig) { this.v = orig.v.Clone(); this.dv = orig.dv.Clone(); }

    /// <summary>
    /// Constructor without partial derivative
    /// </summary>
    /// <param name="v"></param>
    public MultivariateDual(V v) { this.v = v.Clone(); this.dv = new AlgebraicSparseVector<object, V>(); }

    /// <summary>
    /// Constructor for multiple partial derivatives
    /// </summary>
    /// <param name="v"></param>
    /// <param name="keys"></param>
    /// <param name="dv"></param>
    public MultivariateDual(V v, object[] keys, V[] dv) { this.v = v.Clone(); this.dv = new AlgebraicSparseVector<object, V>(keys, dv); }

    /// <summary>
    /// Constructor for a single partial derivative 
    /// </summary>
    /// <param name="v"></param>
    /// <param name="key"></param>
    /// <param name="dv"></param>
    public MultivariateDual(V v, object key, V dv) { this.v = v.Clone(); this.dv = new AlgebraicSparseVector<object, V>(new[] { key }, new[] { dv }); }

    /// <summary>
    /// Constructor with a given value and gradient. For internal use.
    /// </summary>
    /// <param name="v">The value (not cloned).</param>
    /// <param name="gradient">The gradient (not cloned).</param>
    internal MultivariateDual(V v, AlgebraicSparseVector<object, V> gradient) { this.v = v; this.dv = gradient; }

    public MultivariateDual<V> Clone() { return new MultivariateDual<V>(this); }

    public MultivariateDual<V> Zero => new MultivariateDual<V>(Value.Zero, Gradient.Zero);
    public MultivariateDual<V> One => new MultivariateDual<V>(Value.One, Gradient.Zero);

    public MultivariateDual<V> Scale(double s) { v.Scale(s); dv.Scale(s); return this; }

    public MultivariateDual<V> Add(MultivariateDual<V> a) { v.Add(a.v); dv.Add(a.dv); return this; }
    public MultivariateDual<V> Sub(MultivariateDual<V> a) { v.Sub(a.v); dv.Sub(a.dv); return this; }
    public MultivariateDual<V> Assign(MultivariateDual<V> a) { v.Assign(a.v); dv.Assign(a.dv); return this; }
    public MultivariateDual<V> Mul(MultivariateDual<V> a) {
      // (a(x) * b(x))' = b(x)*a(x)' + b(x)'*a(x);
      var t1 = a.dv.Clone().Scale(v);
      var t2 = dv.Clone().Scale(a.v);
      dv.Assign(t1).Add(t2);

      v.Mul(a.v);
      return this;
    }
    public MultivariateDual<V> Div(MultivariateDual<V> a) { v.Div(a.v); dv.Mul(a.dv.Inv()); return this; }
    public MultivariateDual<V> AssignNeg(MultivariateDual<V> a) { v.AssignNeg(a.v); dv.AssignNeg(a.dv); return this; }
    public MultivariateDual<V> AssignInv(MultivariateDual<V> a) { v.AssignInv(a.v); dv.AssignNeg(a.dv).Scale(v).Scale(v); return this; }   // (1/f(x))' = - f(x)' / f(x)^2

    public MultivariateDual<V> AssignSin(MultivariateDual<V> a) { v.AssignSin(a.v); dv.Assign(a.dv).Scale(a.v.Clone().Cos()); return this; }
    public MultivariateDual<V> AssignCos(MultivariateDual<V> a) { v.AssignCos(a.v); dv.AssignNeg(a.dv).Scale(a.v.Clone().Sin()); return this; }
    public MultivariateDual<V> AssignTanh(MultivariateDual<V> a) { v.AssignTanh(a.v); dv.Assign(a.dv.Scale(v.Clone().IntPower(2).Neg().Add(Value.One))); return this; }     // tanh(f(x))' = f(x)'sech²(f(x)) = f(x)'(1 - tanh²(f(x)))

    public MultivariateDual<V> AssignIntPower(MultivariateDual<V> a, int p) { v.AssignIntPower(a.v, p); dv.Assign(a.dv).Scale(p).Scale(a.v.Clone().IntPower(p - 1)); return this; }
    public MultivariateDual<V> AssignIntRoot(MultivariateDual<V> a, int r) { v.AssignIntRoot(a.v, r); dv.Assign(a.dv).Scale(1.0 / r).Scale(a.v.IntRoot(r - 1)); return this; }

    public MultivariateDual<V> AssignExp(MultivariateDual<V> a) { v.AssignExp(a.v); dv.Assign(a.dv).Scale(v); return this; } // exp(f(x)) = exp(f(x))*f(x)'     
    public MultivariateDual<V> AssignLog(MultivariateDual<V> a) { v.AssignLog(a.v); dv.Assign(a.dv).Scale(a.v.Clone().Inv()); return this; }     // log(x)' = 1/f(x) * f(x)' 

    public MultivariateDual<V> AssignAbs(MultivariateDual<V> a) { v.AssignAbs(a.v); dv.Assign(a.dv).Scale(a.v.Clone().Sgn()); return this; }      // abs(f(x))' = f(x)*f'(x) / |f(x)|  doesn't work for intervals      
    public MultivariateDual<V> AssignSgn(MultivariateDual<V> a) { v.AssignSgn(a.v); dv = a.dv.Zero; return this; } // sign(f(x))' = 0;     

    public MultivariateDual<V> AssignMin(MultivariateDual<V> other) {
      throw new NotImplementedException();
    }

    public MultivariateDual<V> AssignMax(MultivariateDual<V> other) {
      throw new NotImplementedException();
    }
  }
}